/**
 * This file contains an abstract implementation of directed graphs.
 * The class provides iterating over the graph and few algorithms.
 * In particular, an algorithm for finding cycles in graph is implemented.
*/

#ifndef DW_GRAPH_H
#define DW_GRAPH_H

#include <map>
#include <set>
#include <climits>
#include <iostream>

template<class VertexT, class lessT, class VectorT, class MatrixT>
class Graph
{

public:

  typedef std::set<VertexT,lessT> SetVertex;

  struct Node{
    int m_distance;
    SetVertex m_edgesIn;
    SetVertex m_edgesOut;
    MatrixT* m_coordSystem;
    MatrixT* m_invCoordSystem;
    Node() : m_coordSystem(0), m_invCoordSystem(0)
    {}

    ~Node()
    {
      if(m_coordSystem)
        delete m_coordSystem;
      if(m_invCoordSystem)
        delete m_invCoordSystem;
    }
  };


  typedef std::map<VertexT,Node,lessT> Nodes;
  typedef typename std::map<VertexT,Node>::iterator iterator;
  typedef typename std::map<VertexT,Node>::const_iterator const_iterator;

  Nodes& nodes()
  {
    return m_nodes;
  }

  // This function clear an information about computed enclosure of the map.
  // It leaves the vertexes only and remove all edges.
  void clearEdges()
  {
    iterator b=begin();
    while(b!=end())
    {
      b->second.m_edgesIn.clear();
      b->second.m_edgesOut.clear();
      ++b;
    }
  }

  // Clear incomming edges - i.e. the transposed graph is removed to free the memory.
  void clearInEdges()
  {
    iterator b=begin();
    while(b!=end())
    {
      b->second.m_edgesIn.clear();
      ++b;
    }
  }

  // This is an auxiliary routine.
  // It creates the transposed graph. 
  void insertInEdges()
  {
    iterator b = begin();
    while(b!=end())
    {
      typename SetVertex::iterator i = b->second.m_edgesOut.begin();
      while(i!=b->second.m_edgesOut.end())
      {
        this->nodes()[*i].m_edgesIn.insert(b->first);
        ++i;
      }
      ++b;
    }
  }

  // This routine removes from the graph all the vertexes
  // that have empty set of incomming edges.
  // In the context of combinatorial maps, these vertexes do not belong
  // to the invariant set since they are not in the image of other vertexes.
  void eraseAcyclicVertices()
  {
    while(true){
      size_t numberOfNodes = m_nodes.size();
      iterator b = begin();
      while(b!=m_nodes.end())
      {
        iterator current = b;
        ++b;
        if( current->second.m_edgesIn.empty() ) //if there is no edge into current
        {
          typename SetVertex::iterator i = current->second.m_edgesOut.begin();
          while(i!=current->second.m_edgesOut.end())
          {
            m_nodes[*i].m_edgesIn.erase(current->first);
            ++i;
          }
          m_nodes.erase(current);
        }
      }

      if(m_nodes.size()==numberOfNodes)
        break;
      std::cout << "Enclosure consists of: " << m_nodes.size() << " boxes" << std::endl;
    }
  }

  // this is the main procedure that finds cycles in the graph
  void findCycles(std::vector<VertexT> result[], int depth)
  {
    int i = static_cast<int>(nodes().size());
    // we traverse the graph 
    iterator b=begin();
    #pragma omp parallel shared(b)
    {
      #pragma omp for schedule(dynamic)
      for(int k=0;k<i;++k)
      {
        iterator it;
        #pragma omp critical
        {
          it = b;
          ++b;
        }
        // here we call the main routine that verifies
        // if the vertex pointed by iterator 'it'
        // belongs to a cycle 
        findCycles(it,result,depth);
      }
    }
  }

  // this subroutine checks if the vertex pointed by 'source'
  // bolongs to a cycle of the lenght <= depth
  void findCycles(iterator source, std::vector<VertexT> result[], int depth)
  {
    SetVertex temp = source->second.m_edgesOut;
    SetVertex image;
    bool found = false;
    for(int i=0;i<depth;++i)
    {
      if(i)
      {
        typename SetVertex::iterator b = image.begin(),
                                     e = image.end();

        while(b!=e)
        {
          temp.insert(nodes()[*b].m_edgesOut.begin(),nodes()[*b].m_edgesOut.end());
          ++b;
        }
      }
      image.clear();
      image.swap(temp);

      #pragma omp critical
      {
        if(image.find(source->first)!=image.end())
        {
          result[i].push_back(source->first);
          found = true;
        }
      }
      if(found)
        return;
    }
  } // end of findCycles

  Nodes m_nodes;

  iterator begin() {
    return m_nodes.begin();
  }

  const_iterator begin() const {
    return m_nodes.begin();
  }

  iterator end() {
    return m_nodes.end();
  }

  const_iterator end() const {
    return m_nodes.end();
  }
};

#endif
