/**
 * This file provides the class MapGraph. This class inherites from the class Graph.
 * The class implements the algorithms that allow to compute graph representations of a map.
 * In particular, Algoirthm 1 from the article is implemented here.
 * 
 * The main template parameter of the class is an abstract map 'f' we want to represent.
 * We assume, it provides operators for computing rigorous bounds of f(x).
*/

#ifndef DW_MAPGRAPH_H
#define DW_MAPGRAPH_H

#include <list>
#include <fstream>
#include "Graph.h"

#include "capd/vectalg/lib.h"
#include "capd/vectalg/Vector.hpp"
#include "capd/matrixAlgorithms/floatMatrixAlgorithms.hpp"

typedef capd::ZVector Point;

template<class PointT>
struct less{

  bool operator() (const PointT& x, const PointT& y) const
  {
    for(int i=0;i<x.size();++i)
    {
      if(x[i]>y[i])
        return false;
      if(x[i]<y[i])
        return true;
    }
    return false;
  }
};

template<class MapT,
        class VertexT=Point,
        class lessT=less<VertexT>,
        class VectorT=typename MapT::VectorType,
        class MatrixT=typename MapT::MatrixType
       >
class MapGraph : public Graph<VertexT,lessT,VectorT,MatrixT>
{
public:
  typedef typename VectorT::ScalarType ScalarType;
  typedef VectorT VectorType;
  typedef MatrixT MatrixType;
  typedef Graph<VertexT,lessT,VectorT,MatrixT> VGraph;
  typedef lessT OrderType;
  typedef typename VGraph::SetVertex SetVertex;
  typedef VertexT Vertex;
  typedef typename MapT::DerivativeContainer DerivativeContainer;

  typedef typename VGraph::iterator iterator;
  typedef typename VGraph::const_iterator const_iterator;
  using VGraph::begin;
  using VGraph::end;

  MapGraph(MapT& _f, const Vertex& _resolution)
    : f(_f), resolution(_resolution), denominator(resolutionToDenominator(_resolution))
  {}

  const Vertex& getResolution() const
  {
    return resolution;
  }

  const Vertex& getDenominator() const
  {
    return denominator;
  }

  MapT& getMap()
  {
    return f;
  }

  static Vertex resolutionToDenominator(const Vertex& resolution)
  {
    Vertex result(resolution.dimension());
    for(int i=0;i<resolution.dimension();++i)
      result[i] = (1 << int(resolution[i]));
    return result;
  }

  
  static VectorT vertexToVector(const Vertex& vertex, const Vertex& denominator)
  {
    VectorT result(vertex.dimension());
    for(int i=0;i<vertex.dimension();++i)
      result[i] = ScalarType(int(vertex[i]),int(vertex[i])+1)/int(denominator[i]);
    return result;
  }

  template<class PointVector>
  static void centreOfBox(const Vertex& vertex, const Vertex& denominator, PointVector& result)
  {
    for(int i=0;i<vertex.dimension();++i)
      result[i] = (double(vertex[i])+0.5)/int(denominator[i]);
  }

  // Here is an important member function.
  // Given the resolution, the function computes the set of boxes
  // that have nonempty intersection with the interval vector 'domain'.
  // These boxes are added to the graph G as vertexes.
  void computeCover(VGraph& G, const VectorT& domain)
  {
    int dim = domain.dimension();
    Vertex left(dim), right(dim);
    int numberOfElements = 1;
    int i;
    capd::rounding::DoubleRounding::roundNearest();

    for(i=0;i<dim;++i)
    {
      double L = domain[i].leftBound();
      double R = domain[i].rightBound();
      left[i] = static_cast<int>(floor(L*int(denominator[i])));
      right[i] = static_cast<int>(ceil(R*int(denominator[i])));
      // these numbers can be same! So we must assure that at least one element is added
      if(left[i]==right[i])
        right[i] = right[i]+1;
      numberOfElements *= right[i]-left[i];
    }

    Vertex coeffs = left;
    for(i=0;i<numberOfElements;++i)
    {
      if(G.nodes().find(coeffs)== G.nodes().end())
        G.nodes()[coeffs] = typename VGraph::Node();
      int k = dim;
      do
      {
        --k;
        coeffs[k] = int(coeffs[k])+1;
        if(int(coeffs[k])==int(right[k]))
          coeffs[k] = int(left[k]);
        else
          k=0;
      }while(k);
    }
  }

  // Here is an important member function.
  // Given the resolution, the function computes the set of boxes
  // that have nonempty intersection with the interval vector 'domain'.
  // These boxes are added to the set of vertexes G.
  void computeCover(SetVertex& G, const VectorT& domain)
  {
    int dim = domain.dimension();
    Vertex left(dim), right(dim);
    int numberOfElements = 1;
    int i;
    capd::rounding::DoubleRounding::roundNearest();
    for(i=0;i<dim;++i)
    {
      double L = domain[i].leftBound();
      double R = domain[i].rightBound();
      left[i] = static_cast<int>(floor(L*int(denominator[i])));
      right[i] = static_cast<int>(ceil(R*int(denominator[i])));
      // these numbers can be same! So we must assure that at least one element is added
      if(left[i]==right[i])
        right[i] = right[i]+1;
      numberOfElements *= right[i]-left[i];
    }

    Vertex coeffs = left;
    for(i=0;i<numberOfElements;++i)
    {
      if(G.find(coeffs)== G.end())
        G.insert(coeffs);
      int k = dim;
      do
      {
        --k;
        coeffs[k] = int(coeffs[k])+1;
        if(int(coeffs[k])==int(right[k]))
          coeffs[k] = int(left[k]);
        else
          k=0;
      }while(k);
    }
  }

  // This function creates a cover of all interval vectors from the list 'domain'.
  // The result is stored in the graph represented by 'this' object.
  // The function just calls the previous routine 'createFirstGrid'
  // for each element from the list.
  void computeCover(const std::list<VectorT>& domain)
  {
    this->nodes().clear();
    typename std::list<VectorT>::const_iterator b = domain.begin(), e = domain.end();
    while(b!=e)
    {
      computeCover(*this,*b);
      ++b;
    }
  }

  // As before, but the result is in the argument 'G'.
  void computeCover(SetVertex& G, const std::list<VectorT>& domain)
  {
    G.clear();
    typename std::list<VectorT>::const_iterator b = domain.begin(), e = domain.end();
    while(b!=e)
    {
      computeCover(G,*b);
      ++b;
    }
  }

  // The following function evaluates the map represented by this graph
  // for each vertex from the set 'vec'.
  // For each vertex the combinatorial representation is computed.
  void computeImageOfBoxes(std::vector<Vertex>& vec)
  {
    int counter;
    int unSize = vec.size();

    #pragma omp parallel shared(unSize)
    {
      #pragma omp for private(counter) schedule(dynamic)
      for(counter=0;counter<unSize;++counter)
      {
        VectorT x = vertexToVector(vec[counter],denominator);

        std::list<VectorT> enc;
        try{
          // here we evaluate the abstract map 'f'
          f.eval(x,enc);
        }catch(std::exception& e)
        {
           std::cout << "exception caught!!!" << std::endl;
           std::cout << e.what() << std::endl;
           std::cout << "x=" << x << std::endl;
           exit(0);
        }

        // After successfull evaluation of the map at 'x'
        // the enclosure of f(x) is assumed to be in 'enc'.
        // We compute its cover and add the result to the graph.
        SetVertex imageGrid;
        computeCover(imageGrid,enc);

        #pragma omp critical
        {
          typename SetVertex::iterator b = imageGrid.begin();
          while(b!=imageGrid.end())
          {
            if(this->nodes().find(*b)==this->end())
            {
              this->nodes()[*b] = typename VGraph::Node();
            }
            ++b;
          }
          this->nodes()[vec[counter]].m_edgesOut.swap(imageGrid);
        } // omp critical
      } // for
    } // parallel    
  }

  
  // This is the main loop of the Algorithm 1 from the article.
  // This procedure calls the method 'computeImageOfBoxes'
  // until there are no new vertexes in the image.
  void generateFirstCover(VectorT initVector)
  {
    this->nodes().clear();
    computeCover(*this,initVector);
    int unSize = this->nodes().size();

    do
    {
      std::vector<Vertex> vec(unSize);
      this->getNewBoxes(vec);
      this->computeImageOfBoxes(vec);
      unSize = this->countNewBoxes();
      std::cout << "number of new boxes: " << unSize << std::endl;
    }while(unSize!=0);

    this->insertInEdges();
    this->eraseAcyclicVertices();
  }

  // An auxiliary routine. It traverse the graph and returns the set of vertexes
  // that do not have computed image.
  void getNewBoxes(std::vector<Vertex>& g)
  {
    g.clear();
    g.resize(countNewBoxes());
    
    iterator b=begin();
    int i=0;
    while(b!=end())
    {
      if(b->second.m_edgesOut.empty())
      {
        g[i] = b->first;
        ++i;
      }
      ++b;
    }
  }

  // An auxiliary routine. It traverse the graph and counts the set of vertexes
  // that do not have computed image.
  int countNewBoxes()
  {
    int r=0;
    iterator b=begin();
    while(b!=end())
    {
      if(b->second.m_edgesOut.empty())
        ++r;
      ++b;
    }
    return r;
  }


  void computeImageOfMap()
  {
    std::vector<Vertex> vec;
    this->getNewBoxes(vec);
    this->computeImageOfBoxes(vec);
  }

  void save(const char* fileName) const
  {
    std::ofstream f;
    f.open(fileName);
    f << resolution;
    const_iterator b=this->begin(), e=this->end();
    while(b!=e)
    {
      f << std::endl << b->first;
      typename SetVertex::const_iterator i= b->second.m_edgesOut.begin();

      while(i!=b->second.m_edgesOut.end())
      {
        f << " " << (*i);
        ++i;
      }
      ++b;
    }
    f.close();
  }

  void saveCoordSys(const char* fileName) const
  {
    capd::rounding::DoubleRounding::roundNearest();
    std::ofstream f;
    f.open(fileName);
    f.precision(20);
    const_iterator b=this->begin(), e=this->end();
    while(b!=e)
    {
      f << b->first;
      if(b->second.m_coordSystem)
        f << " " << *(b->second.m_coordSystem);
      ++b;
      if(b!=e)
        f << "\n";
    }
    f.close();
  }

  void loadCoordSys(const char* fileName)
  {
    capd::rounding::DoubleRounding::roundNearest();
    std::ifstream f;
    f.open(fileName);
    int dim = this->getResolution().dimension();
    Vertex temp(dim);
    do
    {
      f >> temp;
      iterator b = this->nodes().find(temp);
      if(b==this->end())
      {
        std::cout << "An error occured when read coordinate systems\n";
        exit(0);
      }
      b->second.m_coordSystem = new MatrixType(dim,dim);
      f >> *(b->second.m_coordSystem);
      b->second.m_invCoordSystem = new MatrixType(capd::matrixAlgorithms::gaussInverseMatrix(*(b->second.m_coordSystem)));
      f.get();
    }while(!f.eof());
    f.close();
  }

  void load(const char* fileName, bool clearGraph=true)
  {
    if(clearGraph)
      this->nodes().clear();
    std::ifstream f;
    f.open(fileName);
    Vertex temp,temp2;
    f >> resolution;
    denominator = resolutionToDenominator(resolution);

    do
    {
      f >> temp;
      iterator i = this->nodes().find(temp);
      if(i==this->nodes().end())
      {
        this->nodes()[temp] = typename VGraph::Node();
        i = this->nodes().find(temp);
      }
      while(f.get()==' ')
      {
        f >> temp2;
        i->second.m_edgesOut.insert(temp2);
        if(f.eof())
          break;
      }
    }while(!f.eof());
    f.close();
  }

private:
  MapT & f;
  Vertex resolution;
  Vertex denominator;
};

#endif
