/**
 * This file implements the standard Smale map.
 * The required methods for computation of values and derivatives are provided.
*/

#include <list>
#include "capd/vectalg/lib.h"


#ifndef DW_SMALEMAP_H
#define DW_SMALEMAP_H

template<class Scalar>
struct NumConst
{
  const static Scalar pi;
};

template<class Scalar>
const Scalar NumConst<Scalar>::pi = interval::pi();

template<>
const double NumConst<double>::pi = M_PI;

class SmaleMap
{
public:
  typedef capd::IVector VectorType;
  typedef capd::IMatrix MatrixType;
  typedef capd::DVector DVector;
  typedef capd::DMatrix DMatrix;

  // internal class - it stores derivatives and provides trivial affine transformation
  class DerivativeContainer : public std::list<MatrixType>
  {
    public:
    MatrixType affineTransformation(const MatrixType& M)
    {
      iterator b = this->begin(), e = this->end();
      MatrixType result = M *(*b);
      ++b;
      while(b!=e)
      {
        result = intervalHull(result,M*(*b));
        ++b;
      }
      return result;
    }

    void affineTransformation(const MatrixType& M, std::list<MatrixType>& result)
    {
      result.clear();
      iterator b = this->begin(), e = this->end();
      while(b!=e)
      {
        result.push_back(M*(*b));
        ++b;
      }
    }
  }; // DerivativeContainer


  // This operator computes rigorous and nonrigorous value of the map.
  // The formula is given explicite.
  template<class VectorT>
  VectorT operator()(VectorT x) const
  {
    capd::rounding::DoubleRounding::roundNearest();
    typedef typename VectorT::ScalarType Scalar;
    Scalar a = 2*NumConst<Scalar>::pi*x[2];
    while(x[2]<0)
    {
      x[2] += 1;
    }
    while(x[2]>=1)
    {
      x[2] -=1;
    }

    VectorT result(3);
    result[0] = x[0]/10 + 0.5*cos(a);
    result[1] = x[1]/10 + 0.5*sin(a);
    result[2] = 2*x[2];

    if(result[2]>=1)
      result[2] -=1;
    return result;
  }

  // Rigorous derivative - explicite formula.
  MatrixType operator[](const VectorType& x) const
  {
    MatrixType D(3,3);
    D.clear();
    interval a = 2*NumConst<interval>::pi*x[2];
    D(1,1) = interval(1)/10;
    D(1,3) = -NumConst<interval>::pi*sin(a);
    D(2,2) = interval(1)/10;
    D(2,3) = NumConst<interval>::pi*cos(a);
    D(3,3) = 2;
    return D;
  }

  // the operator for computing both value and the derivative.
  template<class VectorT, class MatrixT>
  VectorT operator()(const VectorT& x, MatrixT& result) const
  {
    result = (*this)[x];
    return (*this)(x);
  }

  // nonrigorous derivative - explicite formula.
  DMatrix operator[](const DVector& x) const
  {
    capd::rounding::DoubleRounding::roundNearest();
    capd::DMatrix D(3,3);
    D.clear();
    double a = 2*NumConst<double>::pi*x[2];
    D(1,1) = 0.1;
    D(1,3) = -NumConst<double>::pi*sin(a);
    D(2,2) = 0.1;
    D(2,3) = NumConst<double>::pi*cos(a);
    D(3,3) = 2;
    return D;
  }

  // This function computes the image of a map at vector u
  // The value is a list of interval vectors returned in the argument 'result'.
  // The reason is that the map is defined on the torus.
  // Therefore for some parameters the range of a map cannot be represented 
  // as a single interval vector. 
  //
  // The last parameter is required by a general interface but is ignored here.
  void eval(VectorType u, std::list<VectorType>& result,int =1) const
  {
    // compute the value of the map
    u = (*this)(u);
    result.clear();
    // embed it into the torus.
    if(u[2] < 1)
    {
      result.push_back(u);
      return;
    }
    if(u[2]>=1)
    {
      u[2] -= 1;
      result.push_back(u);
      return;
    }
    VectorType u2 = u;
    u2[2].setLeftBound(1);
    u2[2] -= 1.;
    u[2].setRightBound(1);
    result.push_back(u);
    result.push_back(u2);
  }

  // Rigorous computation of derivative - the result is in DerivativeContainer dc. 
  // Last argument ignored here.  
  void eval(const VectorType& u, const MatrixType& localCoord, DerivativeContainer& dc, int )
  {
    dc.clear();
    dc.push_back((*this)[u]*localCoord);
  }
};

#endif
