/**
 * This file contains implementation of the Algorithm 3 from the article.
*/  

#include <iterator>
#include <algorithm>
#include "posDefMatrix.h"

/**
 * The following function verifies the cone condition for one vertex pointed by iterator 'b'.
 * Namely, for each iterator 'k' such that there is an edge from b->k 
 * we verify if the Df^T Q Df - Q is positive definite
 * 
 * Parameters:
 * GraphT G - graph that contains enclosure of an invariant set, and computed coordinate systems.
 *            We also assume that G can compute rigorously derivatives 
 *            - this is in fact an instance of the class MapGraph.
 * 
 * quadForm - since the quadratic form is always Q(x,y) = |x|^2-|y|^2
 *            we encode it as a vector (diagonal of the matrix).
 * 
 * denominator - a vector of integers, tells us how integer coordinates are 
 *               recomputed to real coordinates.
 * 
 * iterator b - vertex from the graph to be checked in this function
 * 
 * maxSubdivisionLevel - the alrogithm can subdivide the set,
 *                       and recompute derivative when the cone condition is not verified.
 *                       This param says what is the maximum subdivision limit.
 *                       maxSubdivisionLevel = 1 means no subdivision allowed.
 * 
 * Output:
 * true/false - the cone conditions are verified or not.
*/  

template<class GraphT, class iterator>
bool verifyConeConditions(
      GraphT& G,
      const typename GraphT::VectorType& quadForm,
      typename GraphT::Vertex& denominator,
      iterator& b,
      const int maxSubdivisionLevel
    )
{
  typedef typename GraphT::VectorType VectorType;
  typedef typename GraphT::MatrixType MatrixType;
  typedef typename VectorType::ScalarType Scalar;
  typedef typename GraphT::Vertex Vertex;

  const Vertex& vertex = b->first;
  MatrixType* coordSys = b->second.m_coordSystem;

  int n=0;
  while(n<maxSubdivisionLevel)
  {
    if(n)
      std::cout << "computing again the derivative with subdivision level n=" << n << std::endl;
    // Here we take interval vector from the vertex
    VectorType u = GraphT::vertexToVector(vertex,denominator);

    // and call the map from the object G to compute derivative of the map.
    // The function eval (see below) is assumed to return an enclosure of the derivative of the map
    // Df(u)*coordSystem
    // and this result is given in DerivativeContainer dc.
    // This is to improve enclosures when we play with Poincare maps.
    // We use some trick to enclose the product 
    // invCoordSys * Df * coordSys 
    // more efficiently.
    typename GraphT::DerivativeContainer dc;
    G.getMap().eval(u,*coordSys,dc,n);

    // given computed derivative we check the cone condition for each edge in the graph
    // the iterator 'i' traverse outgoing edges from the iterator b. 
    bool verified = true;
    typename GraphT::SetVertex::iterator i = b->second.m_edgesOut.begin();

    while( verified && (i!=b->second.m_edgesOut.end()) )
    {
      // take the coordinate system in the image
      MatrixType& invCS = *(G.nodes()[*i].m_invCoordSystem);
      std::list<MatrixType> derEnc;
      // Here is a special function to evaluated the product
      // invCoordSys * Df * coordSys 
      // see the file 'DerivativeContainer.h"
      dc.affineTransformation(invCS,derEnc);

      // after the above routine we have a list of matrices that enclose 
      // the product invCoordSys * Df * coordSys
      typename std::list<MatrixType>::iterator p = derEnc.begin();
      while(verified && (p!=derEnc.end()) )
      {
        // for each element V from the list we compute 
        // V^T Q V - Q in an efficient way, i.e. assuming that 
        // qudratic forms are diagonal.
        MatrixType V = AT_Q_A_minus_Q(*p,quadForm);
        // eventaully we check positive definiteness.
        verified = verified && checkPosDefMatrix(V);
        ++p;
      }
      ++i;
    } // end while i

    if(verified)
      break;
    ++n;
  } // end while n

  // The above loop stops when either verified==true or n==maxSubdivisionLevel
  // The last case means the cone condition is not verified.

  return (n<maxSubdivisionLevel);
}

// ----------------------- verifyConeConditions ------------------

/**
 * The following function verifies the cone condition for all vertexes in the graph. It just calls 
 * the previuos function for each vertex in the graph.
 * 
 * Parameters:
 * GraphT G - graph that contains enclosure of an invariant set, and computed coordinate systems.
 *            We also assume that G can compute rigorously derivatives 
 *            - this is in fact an instance of the class MapGraph.
 * 
 * quadForm - since the quadratic form is always Q(x,y) = |x|^2-|y|^2
 *            we encode it as a vector (diagonal of the matrix).
 * 
 * maxSubdivisionLevel - the alrogithm can subdivide the set,
 *                       and recompute derivative when the cone condition is not verified.
 *                       This param says what is the maximum subdivision limit.
 *                       maxSubdivisionLevel = 1 means no subdivision allowed.
 * 
 * unverifiedFileName  - the name of a file for printing boxes, on which we were not able to check the cone conditions.
 * 
 * compNo - index of computer in the cluster
 * maxComp - total number of computers in the cluster.
 * 
 * The following routine checks the cone condition for all vertexes that have indexes 'i' satisfying 
 * (i mod maxComp) = compNo
 * 
 * When compNo=0 and maxComp=1 this means that all vertexes are checked.  
*/  

template<class GraphT>
void verifyConeConditions(GraphT& G, const typename GraphT::VectorType& quadForm, const int maxSubdivisionLevel, const char unverifiedFileName[], int compNo, int maxComp)
{
  typedef typename GraphT::VectorType VectorType;
  typedef typename GraphT::MatrixType MatrixType;
  typedef typename VectorType::ScalarType Scalar;
  typedef typename GraphT::Vertex Vertex;

  Vertex denominator = GraphT::resolutionToDenominator(G.getResolution());
  int k;
  int s = static_cast<int>(G.nodes().size());
  
  typedef typename GraphT::Nodes::iterator iterator;
  iterator b = G.nodes().begin();

  // here is some preprocess - we put all the iterators to an array
  std::vector<iterator> itTable(s);
  for(k=0;k<s;++k)
  {
    itTable[k] = b;
    ++b;
  }

  int counterOfUnverified = 0;

  // open file stream to print unverified boxes 
  std::ostringstream fn;
  fn << unverifiedFileName << compNo;

  std::ofstream unverifiedFile;
  unverifiedFile.open(fn.str().c_str());

  #pragma omp parallel shared(b,denominator,counterOfUnverified) private(k)
  {
    #pragma omp for schedule(dynamic)
    for(k=compNo;k<s;k+=maxComp)
    { 
      // this is only to see the progress of computations. Can be commented out.
      if(k<1000 || (k % 1000) < maxComp)
      {
        #pragma omp critical
        {
          std::cout << "Current vertex: " << k << ", " << itTable[k]->first << std::endl;
        }
      }
      
      // we call the previous algorithm for the element itTable[k]
      bool verified = verifyConeConditions(G,quadForm,denominator,itTable[k],maxSubdivisionLevel);
      if(!verified)
      {
        #pragma omp critical
        {
          std::cout << "The vertex: " << k << ", " << itTable[k]->first << " is not verified " << std::endl;
          unverifiedFile << " " << itTable[k]->first << std::endl;
          ++counterOfUnverified;
        }
      }
    }
  } // end parallel

  unverifiedFile.close();
  std::cout << "Number of vertexes for which the cone condition could not be verified: " << counterOfUnverified << std::endl;
}
