/**
 * This file implements some auxiliary nonrigorous algorithms like 
 * - Newton search for periodic points
 * - computation of an approximate Jordan basis
 * - normalization of columns of matrices
*/

#include "capd/alglib/capd2alglib.h"

#ifndef DW_NONRIGOROUS_ALGORITHMS
#define DW_NONRIGOROUS_ALGORITHMS

/**
 * This is a nonrigorous routine - Newton method for finding periodic points.
 * 
 * Parameters:
 * MapT& P - a map. We assume it provides operator for computing 
 *           value and derivative of the represented map.
 * VectorT u - seed point for the Newton search. 
 * period - minimal period of the point
 * iter - number of iterations of the Newton method.
 * 
*/ 
template <class MT, class VT, class MapT>
VT nonrigNewton(MapT& P, VT u, int period, int iter)
{
  capd::rounding::DoubleRounding::roundNearest();
  for(int i=0;i<iter;++i)
  {
    MT D = MT::Identity(u.dimension());
    VT x = u;
    for(int j=0;j<period;++j)
    {
      MT temp;
      x = P(x,temp);
      D = temp*D;
      // out of domain - this is specific for the problems we solved
      if(x.euclNorm()>5) 
        return u;
    }

    u = u - capd::matrixAlgorithms::gauss(D-MT::Identity(u.dimension()),x-u);
    // out of domain - this is specific for the problems we solved
    if(u.euclNorm()>5)
      return u;
  }
  return u;
}


/**
 * This function reads vectors from the given file and put them to the array table.
 * The function is used for example to read cycles found in the graph or
 * candidates for periodic points. Therefore it is a template function.
*/
template<class VectorT>
void loadVectors(std::vector<VectorT> table[], const char filename[])
{
  std::ifstream f;
  f.open(filename);
  VectorT temp;
  int i=0;
  do
  {
    if(f.peek()!='\n')
    {
      do{
        f >> temp;
        table[i].push_back(temp);
        if(f.eof())
          break;
      }while(f.get()==' ');
    }else
      f.get();
    ++i;
  }while(!f.eof());
  f.close();
}


/**
 * Nonrigorous routine. Given a map 'f' it computes the derivative of 'period' iteration
 * of 'f' at 'u'.
*/
template<class MatrixT, class MapT, class VectorT> 
MatrixT computeDerivative(MapT& f, VectorT& u, int period)
{
  MatrixT A = MatrixT::Identity(u.dimension()),
          temp(u.dimension(),u.dimension());
  for(int i=0;i<period;++i)
  {
    u = f(u,temp);
    A = temp*A;
  }
  return A;
}


/**
 * Function normalizes columns of a given martix. 
*/
template<class MatrixT>
void normalizeColumns(MatrixT& A)
{
  for(int j=0;j<A.numberOfColumns();++j)
  {
    A.column(j).normalize();
  }
}


/**
 * Function normalizes some columns of a given martix. 
 * The columns to be normalized have indexes j,...,k-1
*/
template<class MatrixT>
void normalizeColumns(MatrixT& A, int j, int k)
{
  for(;j<k;++j)
  {
    A.column(j).normalize();
  }
}


/**
 * Nonrigorous routine. It computes an approximate Jordan basis of a matrix. 
 * 
 * Parameters:
 * MatrixT A - a matrix that has to be diagonalized
 * 
 * Output:
 * rVec - contains an approximate real parts of eigenvalues
 * true/false - if the matrix have only real eigenvalues or not.
*/
template<class MatrixT>
void computeCoordSystem(const MatrixT& A, MatrixT& rVec)
{
  int dim = A.numberOfRows();
  typename MatrixT::RowVectorType rV(dim), iV(dim);  // real and imaginary parts of eigenvalues
  MatrixT iVec(dim,dim); // real and imaginary parts of eigenvectors
  capd::alglib::computeEigenvaluesAndEigenvectors(A,rV,iV,rVec,iVec);

  // sort eigenvalues
  for(int j=0;j<dim-1;++j)
  {
    for(int k=j+1;k<dim;++k)
    {
      if(fabs(rV[j]) < fabs(rV[k]))
      {
        typename MatrixT::ScalarType t = rV[j];
        rV[j] = rV[k];
        rV[k] = t;
        typename MatrixT::RowVectorType temp = rVec.column(j);
        rVec.column(j) = rVec.column(k);
        rVec.column(k) = temp;

        temp = iVec.column(j);
        iVec.column(j) = iVec.column(k);
        iVec.column(k) = temp;
      }
    }
  }

  // if eigenvalue is complex then take real and imag parts as eigenvectors
  for(int i=0;i<rVec.numberOfColumns();++i)
  {
    if(iV[i]!=0)
    {
      rVec.column(i) = iVec.column(i);
      ++i;
    }
  }

  // eventually normalize columns
  normalizeColumns(rVec);
}

/**
 * This routine computes an approximate Jordan basis of
 * Df^period(u). This is nonrigorous routine.
*/
template<class MatrixT, class MapT, class VectorT>
void computeCoordSystem(MapT& f, VectorT u, int period, MatrixT& result)
{
  // first compute derivative of f^period(u)
  MatrixT A = computeDerivative<MatrixT>(f,u,period);
  //and call the above function to compute Jordan basis of A.
  computeCoordSystem(A,result);
}

/**
 * Nonrigorous routine. It finds an integer coefficients of a box
 * that contains a point u.
*/
template<class VectorT, class Vertex>
Vertex vectorToVertex(const VectorT& u, const Vertex& denominator)
{
  Vertex result(u.dimension());
  for(int i=0;i<u.dimension();++i)
    result[i] = static_cast<int>(floor(u[i]*int(denominator[i])));
  return result;
}

/**
 * For each interval from the range [b,e)
 * an approximate midpoint is computed and inserted to the range pointed by 'i'.
 * This is mainly to convert interval vectors/matrices to real vectors/matrices.
*/
template<class iterator1, class iterator2>
void intervalToDouble(iterator1 b, iterator1 e, iterator2 i)
{
  while(b!=e)
  {
    *i = b->mid().leftBound();
    ++b;
    ++i;
  }
}

/**
 * Given a range [b,e) of reals the routine copy these numbers to the range pointed by 'i'.
 * This is mainly for conversion of real vectors/matrices to interval vectors/matrices.
 * 
*/
template<class iterator1, class iterator2>
void doubleToInterval(iterator1 b, iterator1 e, iterator2 i)
{
  while(b!=e)
  {
    *i = *b;
    ++b;
    ++i;
  }
}

#endif
