#include "nonrigorous-algorithms.h"

#ifndef DW_PERIODIC_POINTS
#define DW_PERIODIC_POINTS

// ----------------------- findCycles -------------------------------------
/**
 * The main algorithm for finding cycles in graph is implemented
 * as a method of class Graph.
 * The following function just take the result of this algorithm and write
 * the output to a file in some format.
 * Also some messages are printed to the screen.
 * 
 * Parameters:
 * Grapht G - a graph that can represent a map.
 * enclosureFilename - a file that contains computed enclosure of an attractor
 * cyclesFilename - file to which the function writes the cycles
 * depth - the maximal lendth of a cycle we are searching for.
*/
template<const int depth, class GraphT>
void findCycles(GraphT& G, const char enclosureFilename[], const char cyclesFilename[])
{
  typedef typename GraphT::Vertex Vertex;

  // We read graph from disc.
  G.load(enclosureFilename);
  std::cout << "number of nodes: " << G.nodes().size() << "\n";

  // declare an array that will contain cycles found.
  std::vector<Vertex> cycles[depth];

  // call the method for finding cycles
  G.findCycles(cycles,depth);

  // and write computed cycles to a file
  std::ofstream cyc;
  cyc.open(cyclesFilename);
  for(int k=0;k<depth;++k)
  {
    Vertex denominator = GraphT::resolutionToDenominator(G.getResolution());
    std::cout << "number of loops of period " << (k+1) << ": " << cycles[k].size() << std::endl;
    for(int j=0;j<static_cast<int>(cycles[k].size());++j)
    {
      cyc << cycles[k][j];
      if(j<static_cast<int>(cycles[k].size())-1)
        cyc << " ";
    }
    if(k<depth-1)
      cyc << std::endl;
  }
  cyc.close();
}

// ----------------------- refineCycles -------------------------------------
/**
 * This is nonrigorous routine. It tries to find a periodic point given a cycle in the graph.
 * The output points are written to a file.
 * 
 * Parameters:
 * Graph G - graph that represents a map.
 * enclosureFilename - a file that contains enclosure of the attractor.
 * cyclesFilename - a file that contains cycles found in the graph
 * perPointsFilename - a file to which the output will be written.
*/
template<const int depth, class VectorT, class MatrixT, class GraphT>
void refineCycles(GraphT& G,
                  const char enclosureFilename[],
                  const char cyclesFilename[],
                  const char perPointsFilename[]
                 )
{
  typedef typename GraphT::Vertex Vertex;

  // L will contain the periodic points that the algorithm will find.
  std::vector<VectorT> L[depth];
  std::ifstream s;
  s.open(enclosureFilename);
  Vertex resolution;
  s >> resolution;
  s.close();
  Vertex denominator = GraphT::resolutionToDenominator(resolution);
  int dim = resolution.dimension();

  // read cycles from disc
  std::vector<Vertex> cycles[depth];
  loadVectors(cycles,cyclesFilename);

  std::ofstream outF;
  outF.open(perPointsFilename);
  outF.precision(20);

  // for each vertex that is in some cycle we proceed as follows.
  for(int k=0;k<depth;++k)
  {
    #pragma omp parallel shared(cycles,denominator)
    {
      #pragma omp for schedule(dynamic)
      for(int j=0;j<static_cast<int>(cycles[k].size());++j)
      {
        VectorT u(dim);
        // take a centre of the box.
        GraphT::centreOfBox(cycles[k][j],denominator,u);
        // perform Newton search - file nonrigorous-algorithms.h
        u = nonrigNewton<MatrixT>(G.getMap(),u,k+1,20);

        VectorT p = u;
        if(p.euclNorm()>5) // out of domain, it is not a periodic point
          continue;

        int s;
        bool isNew=true;
        for(s=0;s<k+1;++s) // lower period than k+1
        {
          u = G.getMap()(u);
          if(s<k && (u-p).euclNorm()<1e-15) // if f^s(u) is very close to u it is not period k point.
            isNew = false;
        }
        if((p-u).euclNorm()>1e-10) // divergent newton method
          continue;
        #pragma omp critical
        {
          for(s=0;s<static_cast<int>(L[k].size());++s)
          {
            // if we already found a point from different seed that is very close to this one
            // then skip it.
            if((L[k][s]-p).euclNorm()<1e-15) 
              isNew = false;
          }
          // if it is really new point write it to the file.
          if(isNew)
          {
            L[k].push_back(p);
            outF << " " << p;
            std::cout <<  (k+1) << "-periodic point found:" << p << std::endl; 
          }
        } // omp critical
      } // for
    } // omp parallel
    if(k<depth-1)
      outF << std::endl;
  }

  outF.close();
}

// ----------------------- computeCoordSystemsAtFixPoints -------------------
/**
 * This is rather obvious routine. For each vertex that contains a k-periodic point 'u'
 * we compute coodinate system at this vertex as a Jordan basis od Df^k(u).
 * 
 * Parameters:
 * GraphT G - graph that encloses a map.
 * perPointsFilename - a file that contains numerical found periodic points.
 * enclosureFilename - a file that contains enclosure of the attractor
 * 
 * Output:
 * SetVertex checked - a set of vertexes that will have assigned a coordinate system in this routine.
*/
template<const int depth, class VectorT, class MatrixT, class GraphT>
void computeCoordSystemsAtFixPoints(
    GraphT& G,
    const char perPointsFilename[],
    const char enclosureFilename[],
    typename GraphT::SetVertex checked[]
  )
{
  typedef typename GraphT::VectorType VectorType;
  typedef typename GraphT::MatrixType MatrixType;
  typedef typename GraphT::Vertex Vertex;

  // Read periodic points from disc.
  std::vector<VectorT> L[100];
  loadVectors(L,perPointsFilename);

  // Read enclosure of the attractor from disc.
  G.nodes().clear();
  G.load(enclosureFilename);

  Vertex resolution = G.getResolution();
  Vertex denominator = GraphT::resolutionToDenominator(resolution);

  int dim = resolution.dimension();
  int i,j;

  // For each periodic point proceed:
  for(i=0;i<depth;++i)
  {
    #pragma omp parallel shared(L)
    {
      #pragma omp for schedule(dynamic)
      for(j=0;j<(int)L[i].size();++j)
      {
        // find a box that contains periodic point
        Vertex vertex = vectorToVertex(L[i][j],denominator);
        typename GraphT::iterator it = G.nodes().find(vertex);

        // if the box is out of enclosure, skip it.
        if(it==G.end())
          continue;
        
        // If periodic point belongs to a vertex that has already computed coordinate system 
        // from another periodic point then we skip it.
        if(it->second.m_coordSystem!=0)
          continue;

        // compute Jordan basis - in file nonrigorous-algorithms.h
        MatrixT df(dim,dim);
        computeCoordSystem(G.getMap(),L[i][j],i+1,df);
        
        // copy Jordan matrix to an interval matrix and rigorous inverse of it.
        MatrixType cs(dim,dim), invcs(dim,dim);
        std::copy(df.begin(),df.end(),cs.begin());
        invcs = capd::matrixAlgorithms::gaussInverseMatrix(cs);

        #pragma omp critical
        {
          std::cout << "computing coordinates for: " <<  L[i][j] << std::endl;
          it = G.nodes().find(vertex);
          // Assign coordinate system. 
          // In parallel version we must check again if the vertex has already computed coordinates.
          if(it->second.m_coordSystem==0)
          {
            it->second.m_coordSystem = new MatrixType(cs);
            it->second.m_invCoordSystem = new MatrixType(invcs);
            checked[0].insert(vertex);
          }
        } // omp critical
      } // for j
    } // omp parallel
  } // for i
  std::cout << "Coordinates at periodic points computed\n\n";
}

#endif
