#include <ctime>
#include <cstdlib>
#include <list>
#include <vector>
#include <omp.h>

// here we include all the implemented algorithms
#include "MapGraph.h"
#include "coordSystems.h"
#include "coneConditions.h"

// here is the class that represents the Smale map.
#include "SmaleMap.h"
typedef MapGraph<SmaleMap> SmaleMapGraph;

static const char sEnclosureFilename[] = "smale.dat";
static const char sCycleFilename[] = "sCycles";
static const char sPerPointsFilename[] = "sPerPoints";
static const char sCoordSysFilename[] = "sCoordSys";
static const int sDepth = 3;

using namespace capd;
using std::cout;
using std::endl;


void smaleExample()
{
  const int res = 4; // this is the minimal value for which the cone condition can be verified.
  int resData[]={res,res,res};
  SmaleMapGraph::Vertex resolution(3,resData);

  SmaleMap s;
  // we define the graph that enclose the Smale map 's' with the resolution 'resolution'
  SmaleMapGraph G(s,resolution);

  // we take some initial point
  IVector initPoint(3);
  initPoint[0] = interval(0.1,0.10000001);
  initPoint[1] = interval(0.5,0.50000001);
  initPoint[2] = interval(0.5,0.50000001);

  // Call of the Algorithm 1 to enclose the attractor
  cout << "-------- Enclosing of attractor (Algorithm 1) -----------" << endl;
  G.generateFirstCover(initPoint);
  G.save(sEnclosureFilename);
  cout << "done" << endl << endl; 

  cout << "----------------- Detecting cycles in the graph -------------" << endl;
  // Call of the algorithm for detecting cycles in graph - in the file periodicPoints.h
  findCycles<sDepth>(G,sEnclosureFilename,sCycleFilename);
  cout << "done" << endl << endl; 

  std::cout << "----------------- Detecting fixed points -------------" << endl;
  // Call of the algorithm for refining cycles to periodic points - in the file periodicPoints.h
  refineCycles<sDepth,DVector,DMatrix>(G,sEnclosureFilename,sCycleFilename, sPerPointsFilename);
  cout << "done" << endl << endl; 

  std::cout << "----------------- Computing of coordinate systems -------------" << endl;
  // In the following routine we compute cooordinate systems at the periodic points,
  // then we call Algorithm 2 for spreading coordinate system from periodic points.
  // Implementation in the files periodicPoints.h and coordSystems.h.
  computeCoordSystems<sDepth,DVector,DMatrix>(G,sPerPointsFilename, sEnclosureFilename);
  G.saveCoordSys(sCoordSysFilename);
  cout << "done" << endl << endl; 

  // This is quadratic form - one positive and two negative values
  IVector Q(3);
  Q[0] = 1;
  Q[1] = -1;
  Q[2] = -1;

  std::cout << "----------------- Verification of the cone condition (Algorithm 3) -------------" << endl;
  // Call of the Algorithm 3 for verification of the cone condition - in coneCondition.h.
  // the parameters 1,0,1 are not important - they specify the subdivision levels, if necessary
  // number of a computer in cluster and total number of computers in cluster.
  verifyConeConditions(G,Q,1,"unverified",0,1);
}

int main(int argc, char** argv)
{
  try{
    smaleExample();
    cout << "Finished!\n";
  } catch(std::exception& e)
  {
    cout << "\nException caught!\n" << e.what() << endl;
  }

  return 0;
}
